## ############### 
## The NT specific subroutines for SPEC95
##
# $Id: nt.pl,v 7.3 1996/09/04 19:09:50 bsmith Exp $


## ##############
## SUB                        PWD
## ##############

sub pwd {
    my $pwd;
    chop($pwd = `chdir`);
    $pwd =~ s#^[A-Za-z]:##;	# remove any drive designators
    $pwd =~ s#\\#/#g;		# replace the backslashes with forward ones
    $pwd = &joinpaths($pwd);	# ensure correct format
}

## ##############
## SUB                      FIND_TOP
## ##############

sub find_top {	## sub for finding the top of the SPEC tree

    ## if "$SPEC/bin/$program" is exectuable

    if ( ! -f &joinpaths($ENV{'SPEC'}, "bin", "$program")) {
	$dir = $pwd;	## using the global $pwd
	$dir =~ s#/$##;	## delete a trailing slash if there
	while ($dir) {	## iterate while a $dir exists

            ## break loop if there is a $dir/bin/$program
	    ## this is the key that a spec tree has been found
            last if -f &joinpaths($dir, "bin", "$program");

	    ## otherwise delete everything beyond the last slash in $dir
 	    $dir =~ s#^[A-Za-z]\:##; ## remove and drive designators
	    $dir =~ s#/[^/]*$##;
	}
	if (!$dir) {
	    die "Error: Can't find TOP of spec tree\n";
	}
    } else {	## there is a $SPEC/bin/$program
	    $dir = $ENV{'SPEC'};
    }

    ## set the associative array to this value

    $spec{'top'} = $dir;
}


## ##############
## SUB                       JOINPATHS
## ##############

sub joinpaths {	## sub for concantenating file path elements
                ## using VMS directory notation or UNIX notation
                ## result is always returned as absolute path UNIX style
                ## with device name as part of the path
    my(@args) = @_;
    my($result);

    $result = join('/',@args);
    $result =~ s#\\#/#g;          # replace any backslashes
    $result =~ s#[A-Za-z]\:##;    # remove any drive designators if there
    $result =~ s#^#/# unless ($result =~ m#^/#); ## insert a leading slash
                                                 ## if needed
    $result;
}

## ##############
## SUB                      SETUPENV
## ##############
## set up environment.. OS specific

sub setupEnv {
    $ENV{'SPEC'} = $spec{'top'};
    $ENV{'SPEC'} =~ s/\//\\\\/g;
    $decap = substr($spec{'top'},2);  #chop the C:
    $ENV{'PATH'} = "$spec{'top'}/bin;$spec{'top'}/bin/lib;$ENV{'Path'}";
    unshift(@INC, 
       &joinpaths($spec{'top'},"bin"), 
       &joinpaths($spec{'top'},"bin","lib")
       );

## we assume the DynaLoader isn't working on VMS, but if it is then uncomment
## the next line
# use Spec;

}

## ##############
## SUB                      SETPROCGRP
## ##############
## OS Specific

# Set up a process group if we can
## the use of eval is a trick to trap errors in a subroutine. It
## forces the Perl interpretter to evaluate the string as if it were
## Perl code (which we hope it is!). The question here is, setpgrp
## implemented in the version of Perl that we are running? This is
## OS specific and should be tested before being run !!!!!!!!!!!

sub setProcGrp {
    # NT doesn't do this stuff
    1;
}

## ##############
## SUB                     SETOSDEFAULTS
## ##############
## set OS specific defaults in the %user and %spec structures

sub setOSdefaults {
    $spec{'make'} = "specmake SHELL=cmd";	# Name of make executable
    $user{'shell'} = "C:\\WINNT35\\system32\\cmd.exe"; # default shell to run
    $user{'endian'} = 0x1234;	# rather than trying to run Spec::endian;
}

## ############
## SUB                   LIST_FILES
## ############

sub list_files {
    my (@list, $file);
    local (*DIR);
    for $file (@_) {
	if (!-l $file && -d $file) {
	    opendir(DIR, $file);
	    my (@files) = readdir(DIR);
	    closedir(DIR);
	    @files = grep (!m/^\.\.?$/, @files);
	    @files = grep (s#^#$file/#, @files);
	    push (@list, &'list_files(@files));
	} elsif (-f $file) {
	    push (@list, $file);
	}
    }
    return (@list)?@list:undef;
}

## ###########
## SUB                            RUN_BENCHMARK
## ##########
## the UNIX version uses fork() and wait() but other OS do not 
## spawn processes in the same way and so this subroutine must be OS
## specific

sub OS_run {
    my ($me, $size, $tune, $ext, $mach, $users, $fdocommand) = @_;
    my ($return_code) = 1;
    my ($i, @start_time, @stop_time, @diff_time, $name, $pid, $action, $rc);
    my (@exitval, @num_targets, @todo, $run_num);
    my ($str, @temp, @temp2, @submit_commands, %config, $submit);
    my $benchtop  = $me->path;
    my $benchname = $me->name;
    my $benchnum  = $me->num;
    my $benchmark = $me->benchmark;
    my $spectop = $'spec{'top'};
    my $count = 0;
    my $command = "";
    my $temp;
    my $wait_pid;
    my $noisy = ($'user{'teerunout'} =~ m/^y(es)?/i) &&
		($'user{'teeout'} =~ m/^y(es)?/i);
    ;
    local (*TIMEFILE, *FILE);

    %config = &'build_config($me->benchmark, $tune, $ext, $mach);

    ## for some reason Bill (NT port) removed the sort for the following
    ## instruction ???
    for (sort &'key_match("submit*", \%config)) {
	push (@submit_commands, $config{$_});
    }
    $submit = 0;

    $i = 0;

    ## defined benchmark run variables available to config files
    my (%repl) = (
		    'statement' => \$count,
		    'command'   => \$command,
		    'dir'       => \$dir,
		    'user'      => \$i,
		    'numusers'  => $users,
		    'benchmark' => $benchmark,
		    'benchtop'  => $benchtop,
		    'benchnum'  => $benchnum,
		    'benchname' => $benchname,
		    'spectop'   => $spectop,
		);

# Build a list of what to do in each directory and clobber any output files
    for ($i = 0; $i < $users; $i++) {
	$dir = $me->{'work'}{'dirs'}[$i]; ## a separate work directory for 
	                                  ## each "user"
	chdir $dir;
	unlink $me->outputs($size); ## clean out the output files
	@temp = $me->invoke($size, $ext); ## the raw invocation list for this 
				          ## user and bench
	@temp2 = ();
	$count = 0;
	for (@temp) {
	    $command = $_;

	    ## expand variables and values in the command line
	    if (defined $fdocommand) {
		## substitute in values for variables in $fdocommand
		$command = &'subst(\%repl, $fdocommand);
	    }

	    if (defined $config{'monitor_wrapper'}) {
		## substitute in values for variables in monitor_wrapper
		$command = &'subst(\%repl, $config{'monitor_wrapper'});
	    }

	    if (@submit_commands && $'spec{'mode'} eq "rate") {
		## substitute in values for variables in $submit_command..
		$command = &'subst(\%repl, $submit_commands[$submit]);
	    }

	    push (@temp2, $command); ## prepare for secondary processing
	    $count++;
	} ## end of for(@temp)

	## secondary list processing
	@{$todo[$i]} = @temp2; ## put the commands into a list of lists
	$num_targets[$i] = @temp2;
	&'log(20, "$benchmark user $i\n");
	for (@temp2) {
	    &'log(21, "    $_\n");
	}
	$submit = 0 if ++$submit > $#submit_commands;
    } ## end of for ($i =0; $i < $users; $i++)

# Open up the exitval file now so that if we fail somewhere we at least have
# a partial file in the directory
    $run_num = (&'find_biggest_file($me->work, $me->name."_".$'spec{'timefile'}))[3];
    $name = &'joinpaths($me->work,$me->name."_".$'spec{'timefile'}.".$run_num");
    if (open  (TIMEFILE, ">$name")) {
	print TIMEFILE $me->benchmark, "\n";
    } else {
        &'log(0, "Can't write to '$name': $!\n");
    }

    chdir $'spec{'top'};
    if ($config{'monitor_pre_bench'}) {
	my $command = &'subst(\%repl, $config{'monitor_pre_bench'});
	system $command;
    }
# Start off first job for each directory
# We fork off a job based on what was returned by the 'invoke command.
# If it's an array we pass it to exec as an array so we don't incur a shell
# If it's a scalar we pass it as a scalar, this will invoke the system's
# /bin/sh equivalent.


    @start_time = &spectime; # the proper way
    for ($i = 0; $i < $users; $i++) {
	chdir $me->{'work'}{'dirs'}[$i];
	$action = shift(@{$todo[$i]}); ## grab a command off the list
	$fork_time = time;

	$action =~ tr/\//\\/;
	if (&Win32::CreProc("$ENV{'ComSpec'}", "/c $action", $pid)) {
	    $children{$pid} = $i; ## use as an index to the user number
	    $usedpid{$pid} = 1;	## flag the proces as active

	    ## a bunch of non-critical output
	    if ($noisy) {
		if (ref($action) eq "ARRAY") {
		    print join (" ", @$action), "\n";
		} elsif (ref($action) eq "CODE") {
		    print "perl code\n";
		} else {
		    print $action, "\n";
		}		# 
	    }
	} else {		## there must have been a problem spawning
	    ## signal all existing children
	    kill TERM, keys %children;
	    &'log (0, "Error Spawning Children: $!\n");
	    exit 1;
	}

	## log the activity
	if ($'log_open) {
	    if (ref($action) eq "ARRAY") {
		$str = join (" ", @$action);
	    } elsif (ref($action) eq "CODE") {
		$str = "Perl code forked";
	    } else {
		$str = $action;
	    }
	    &'log (98, "Forked child $pid for user $i at $fork_time: $str\n");
	}
    } ## end of "for ($i = 0; ..."

    ## now it's time to wait on the children
    while (@procKeys = keys(%children)) {

	$wait_pid = &Win32::Wait(@procKeys, $rc);

	$wait_time = time;
	&'log (98, "Child $wait_pid terminated with return code $rc at $wait_time\n") if ($'log_open);
	if ($wait_pid == -1) {
	    &'log (0, "No children, but I think I have these left: ",
	             join (", ", sort { $a <=> $b } keys %children), "\n");
	    exit 1;
	} elsif (defined $children{$wait_pid}) {
	    $dir = $children {$wait_pid};
	    delete $children {$wait_pid};
	    push (@{$exitval[$dir]}, $rc);
	    chdir $me->{'work'}{'dirs'}[$dir];
	    $action = shift(@{$todo[$dir]});
	    next if !defined $action;
	    $fork_time = time;
	    $action =~ tr/\//\\/;
	    if (&Win32::CreProc("$ENV{'ComSpec'}", "/c $action", $pid)) {
		$children{$pid} = $dir;
		$usedpid{$pid} = 1;
	    } else {
		kill TERM, keys %children;
		&'log (0, "Error Forking Children: $!\n");
		exit 1;
	    }
	} else {
	    &'log(98, "Child I didn't spawn reported its death: $wait_pid\n");
	    if ($usedpid{$wait_pid}) {
		&'log(98, "Ok, maybe I spawned it... We'll let it pass\n");
	    } else {
		exit 1;
	    }
	}
	if ($'log_open) {
	    if (ref($action) eq "ARRAY") {
		$str = join (" ", @$action);
	    } elsif (ref($action) eq "CODE") {
		$str = "Perl code forked";
	    } else {
		$str = $action;
	    }
	    &'log (98, "Forked child $pid for user $dir at $fork_time: $str\n");
	}
    }
    @stop_time = &spectime; # a hack
    chdir $'spec{'top'};
    if ($config{'monitor_post_bench'}) {
	my $command = &'subst(\%repl, $config{'monitor_post_bench'});
	system $command;
    }

    ## note that the start time is the one that is outside the loop
    @diff_time = &'sub_time(@stop_time, @start_time);
    &'log (97, "Benchmark Times:\n",
              "  Start:    ", join(", ", @start_time, &'ctime($start_time[0])),
              "  Stop:     ", join(", ", @stop_time, &'ctime($stop_time[0])),
              "  Elapsed:  ", join(", ", @diff_time, &'to_hms(@diff_time)),
	      "\n");

# Write out a file in each directory which contains the exit values of
# each run
    for ($i = 0; $i < $users; $i++) {
	$name = &'joinpaths($me->{'work'}{'dirs'}[$i], $'spec{'exitvals'}.".$run_num");
	if (open  (FILE, ">$name")) {
	    print FILE $me->benchmark, "\n";
	    print FILE join(", ", @start_time, &'ctime($start_time[0]));
	    print FILE join("\n", @{$exitval[$i]}), "\n";
	    if ($num_targets[$i] != @{$exitval[$i]}) {
		print FILE "ERROR: Missing Runs!\n";
		&'log(97, "User $i appears to be missing some runs?\n");
	    }
	    close (FILE);
	} else {
	    &'log(0, "Can't write to '$name': $!\n");
	}
    }

# Write out a file containing the run time of the benchmarks
    print TIMEFILE join(", ", @'bench_time, &'ctime($'bench_time[0]));
    print TIMEFILE join(", ", @start_time, &'ctime($start_time[0]));
    print TIMEFILE join(", ", @stop_time,  &'ctime($stop_time[0]));
    print TIMEFILE join(", ", @diff_time,  &'to_hms(@diff_time)), "\n";
    print TIMEFILE "$size, $tune, $ext, $users\n";
    close (TIMEFILE);

    sleep $config->{'wait_after_run'} if ($config->{'wait_after_run'} > 0);
	
    chdir $'pwd;
}

## ###########
## SUB                            COMPARE_BENCHMARK
## ##########

## Spawns child processes to do the compares of the outputs of a 
## benchmark, then manages the results of these.
## the UNIX version uses fork() and wait() but other OS do not 
## spawn processes in the same way and so this subroutine must be OS
## specific

sub OS_compare {
    my ($me, $size, $users) = @_;
    my ($outdir, $i, @rc, $pid, $date, $name, $exitvals, $error);
    my (%children, $j, $run_num, $target_date, $wait_pid, $name2, $rc);
    my (%rc, %bytype, @timefile, @temp);
    my ($active) = 0;
    local (*FILE);

    $run_num = (&'find_biggest_file($me->work, $me->name."_".$'spec{'timefile'}))[1];
    $name = &'joinpaths($me->work,$me->name."_".$'spec{'timefile'}.".$run_num");
    open (FILE, "<$name");
    @timefile = <FILE>;
    close (FILE);
    $target_date = $timefile[2];
    if (!defined $target_date || !$target_date) {
	push (@rc, "Last run appears to have run improperly: $name has no data!\n");
	$error = 1;
    }
    
    # make sure that all of these runs were done at the same time
    if (!$error) {
	for ($i = 0; $i < $users; $i++) {
	    $name = &'joinpaths($me->{'work'}{'dirs'}[$i], $'spec{'exitvals'}.".$run_num");
	    if (!open (FILE, "<$name")) {
		$rc{$i} = "nofile";
		next;
	    }
	    @temp = <FILE>;
	    close (FILE);
	    $date = $temp[1];
	    if (!defined $date) {
		$error = 1;
		$rc{$i} = "nodate";
	    } elsif ($target_date ne $date) {
		$error = 1;
		$rc{$i} = "date";
	    }
	    chomp(@temp);
	    for ($j = 2; $j < @temp; $j++) {
		if ($temp[$j] != 0) {
		    $rc{$i} = "exitval";
		}
	    }
	}
    }

    if (!$error) {
	for ($i = 0; $i < $users; $i++) {
	    next if defined $rc{$i};
	    &'log (98, "Compare user $i output\n");
	    $error = 0;
# Find information about the run that this was associated with
	    $name = &'joinpaths($me->{'work'}{'dirs'}[$i], $'spec{'comparename'}.".$run_num");
	    if (!open (FILE, ">$name")) {
		&'log(0, "Can't write to '$name': $!\n");
		exit 1;
	    }
	    print FILE $me->benchmark,"\n";
	    print FILE "$date\n";
	    for ($me->outputs($size)) {
		@rc = &'spec_diff(
		    $'spec{'difflines'}, 
		    $me->{'data'}{$size}{$'spec{'outputdir'}}{$_}, 
		    &'joinpaths($me->{'work'}{'dirs'}[$i], $_),
		    1, 
		    $me->{'tolerance'}{'absolute'}, 
		    $me->{'tolerance'}{'relative'}, 
		    );
		if (@rc > 0) {
		    $error = 1;
		    print FILE "ERROR: Mismatch in '$_'\n";
		    print FILE @rc;
		    print FILE "\n";
		}
	    }
	    close (FILE);

	    &'log (98, "Compare yielded $error\n");

            $rc{$i} = "compare" if $error;
	}
    }
    
    for (sort { $a <=> $b } keys %rc) {
	push (@{$bytype{$rc{$_}}}, $_);
    }
    if (@{$bytype{'nofile'}}) {
	push (@rc, "The following users did not have a $'spec{'timefile'}.$run_num file: ".join (", ", @{$bytype{'nofile'}})."\n");
    }
    if (@{$bytype{'nodate'}}) {
	push (@rc, "The following users' $'spec{'timefile'}.$run_num file did not contain a date: ".join (", ", @{$bytype{'nodate'}})."\n");
    }
    if (@{$bytype{'date'}}) {
	push (@rc, "The following users' $'spec{'timefile'}.$run_num file contained the wrong date: ".join (", ", @{$bytype{'date'}})."\n");
    }
    if (@{$bytype{'exitval'}}) {
	push (@rc, "The following users returned non-zero return codes: ".join (", ", @{$bytype{'exitval'}})."\n");
    }
    if (@{$bytype{'compare'}}) {
	push (@rc, "The following users produced the wrong output: ".join (", ", @{$bytype{'compare'}})."\n");
    }
    $name2 = &'joinpaths ($me->results, (&'find_biggest_file($me->results, $me->name))[2]);
    open (FILE, ">$name2");
    print FILE @timefile;
    if (@rc) {
	&'log(96, @rc);
	print FILE @rc;
    } else {
	print FILE "Successful Run ", $me->benchmark, "\n";
	print FILE @rc;
	&'log(96, "Successful Run ", $me->benchmark, "\n", @rc);
    }
    close (FILE);
    if (@rc) {
	unshift(@rc, "Error during compare see file '$name2'\n");
    }
    (scalar(@rc))?@rc:undef;
}

## ##############
## SUB                      USAGE
## ##############

## because some operating systems do not support case sensitive
## command line interpreters (unlike UNIX), the option list
## needs to be OS specific
## Amazingly, Windows-NT passed the command line in a case sensitive
## fashion, and therefore we may duplicate the UNIX optlist

$optList= "Aa:c:e:dDhH?Ii:o:n:rt:s:T:u:Vv:m:fCl:x:";

sub usage { ## inform the user of the command line options
    print <<EOT;
Usage: $0 [options] list_of_benchmarks
List of Options:
Most Frequently Used Options:
       -a action    runspec action to perform [build|setup|run|compare|
       		    report|validate|clean|trash|clobber|nuke]
       -c name      name of configuration file in $SPEC/config to use
       -e ext       binary name extension/suffix to use; overrides config file
       -h           display runspec usage message (this message)
       -i input     SPEC input files to use [test|train|ref]; Default: ref
       -m mach      machine configuration to use; overrides config file
       -n num       Number of times to run each benchmark (must be at least
       		    3 for "official" reporting runs)
       -o type      type of output report to produce [asc|html|ps|raw|all)
       -r           run throughput metric (SPECrate) instead of speed metric
       -T tune      type of compilation tuning [base|peak|all]; overrides
       		    config file
       -u copies    comma separated list of # of concurrent copies for SPECrate

Less Frequently used options:
       -C	    Do not remove *.o before building (*USE WITH CAUTION*)
       -d           explicitly delete benchmark working sub-directories; 
       		    setup, run and compare by default use existing directory
       -D           delete prexisting compiled binaries for current
       		    configuration 
       -f           Cause STDOUT to be unbuffered
       -H           Heartbeat during run (will not work on all systems)
       -I           Ignore errors; do not exit when error occurs
       -l num       # of benchmark result files to search back for report data
       -s os        operating system (default is 'unix')
       -v level     set output/log verbosity level (default = 2, range=0-100)
       -V           print version number
       -x           debugging level mask (base 2)

list_of_benchmarks may be 'all', 'int', 'fp', a list of benchmarks desinated 
by number.name, just the number, or enough characters of their full name 
to be unique.  

EOT
    exit 1;
}
## ###########
## SUB            COMMAND_BENCH
## ###########

sub command_bench {
    my ($cmd, $dir, $outn, $repl) = @_;
    ## $cmd   -- initially, the unexpanded command string; eventually the whole
    ## $dir   -- the directory in which the command is to be executed
    ## $outn  -- basename for output and errors files
    ## $repl  -- the array of replacement variables and values

    my $errname = "$outn.err";
    my $outname = "$outn.out";
    my $rc;

    ## move to the directory
    chdir $dir;

    ## expand the $cmd with the variables in %{$repl}
    $cmd = &subst_with_environ ($repl, $cmd) if defined $repl;

    ## ready -- make a log entry if required
    &log (100, "Issuing command '$cmd'\n");

    ## set -- add tee to $cmd, if required
    if ($outn ne "") {
	unlink $errname, $outname;
	if ($'user{'teeout'} =~ m/^y(es)$/i) {
	    $cmd .= " 2> $errname | tee $outname";
	} else {
	    $cmd .= " 2> $errname > $outname";
	}
    }

    ## give user some indication of what is happening if she is
    ## is about to get some tee output
    if ($'user{'teeout'} =~ m/^y(es)?$/i) {
	    print $cmd, "\n";
    }

    ## go -- this is it.. issue the command and grap the result
    $rc = system $cmd;

    ## if an output basename is specified
    if ($outn ne "") {
	if (-s $outname > 0) { ## and if the output file exists
                               ## put the contents into the log file
	    &log(100, "Output from '$cmd':\n", 
		    &read_file(&joinpaths($dir,$outname)));
	}
	if (-s $errname > 0) { ## and if the error file already exists
                               ## put the contents into the log file
	    &log(100, "Error  from '$cmd':\n", 
		    &read_file(&joinpaths($dir,$errname)));

	    ## and if there is a tee involved, display the contents of the
            ## error file
	    if ($'user{'teeout'} =~ m/^y(es)?$/i) {
		    print &read_file(&joinpaths($dir,$errname));
	    }
	}
    }

    ##                  ERROR HANDLING
    if ($rc) { ## if there is a non-zero result from the $cmd
	if ($rc == $spec{'sigint'}) { ## the command was interrupted
	    &log(0, "Exit on user ctrl-c\n"); ## not always true -- see stty
	    exit 1; ## drop dead with poison arrow
	}

	## otherwise, we have a problem issuing the command to system
        ## so grab the system error message from $! (provided $error points
        ## to something)
	$$error = "$!" if defined $error && ref $error;

	## $temp holds the path for the error file
	$temp = &joinpaths($dir, $errname);
	&log(0, "Error with '$cmd': check file '$temp'\n");


	## read the error file into a list
	open (FILE, "<".&joinpaths($dir, $errname));
	while (<FILE>) {
	    push (@list, $_);
	    if (@list > ($spec{'difflines'} + 10) * 2) {
		splice(@list, 0, @list - $spec{'difflines'});
	    }
	}
        ## remove the "difflines"
	splice(@list, 0, @list - $spec{'difflines'});

        ## spit out ther resulting error
	&log(0, \@list);

	close (FILE);
    }
    ## ??? shouldn't there be an $rc here to insure the system result returned
}


###########
## SUB             spectime
###########

sub spectime {			# a wrapper for Win32::SpecTime 
    my($sec,$usec) = 0;
    &Win32::SpecTime($sec,$usec);
    $usec *= 1000;      # We need microseconds; we get milliseconds
    return($sec, $usec);
}

1;



