package Spec::Format::raw;

#
#  raw.pl - produces RAW output
#  (c) Standard Performance Evaluation Corporation 1995
#
#  Author:  Christopher Chan-Nui
#
# $Id: raw.pl,v 1.6 1995/08/17 00:36:55 channui Exp $

sub new {
    bless {};
}

sub extension {
    "";
}

sub format {
    my($me, $results) = @_;
    my(@output);
    my ($pattern, $gpattern, $bn, $f_name, $f_ratio, $f_ref, $f_time);
    my $i;
    my $temp;
    my $tune;
    my %ref_printed;

    push (@output, "SPEC95RAW1.11");

    for $tune (@'valid_tunes) {
	for $bench (sort keys %{$results->{$tune}}) {
	    if ($bench =~ m/^\d{3}\.\S+$/) {
		$f_name  = $_;
		$f_ref   = $results->{$bench}{'reference'};
		$f_ref   = "N/A" if !defined $f_ref;
		$f_best  = $results->{$tune}{$bench}{'time'};
		$f_best   = "N/A" if !defined $f_best;
		if (defined $results->{$tune}{$bench}{'index'}) {
		    if (ref ($results->{$tune}{$bench}{'index'}) eq "ARRAY") {
			$f_index  = join (" ", 
					@{$results->{$tune}{$bench}{'index'}});

		    } else {
			$f_index  = $results->{$tune}{$bench}{'index'};
		    }
		} else {
		    $f_index   = "N/A";
		}
		$f_copies  = $results->{$tune}{$bench}{'copies'};
		$f_copies   = "N/A" if $f_copies == 0;
		if (defined $results->{$tune}{$bench}{'ratio'}) {
		    $f_ratio = $results->{$tune}{$bench}{'ratio'};
		} else {
		    $f_ratio = "N/A";
		}
		if (!$ref_printed{$bench}) {
		    push (@output, "$bench: reference        $f_ref");
		    $ref_printed{$bench} = 1;
		}

		push (@output, "$bench: $tune ratio       $f_ratio");
		push (@output, "$bench: $tune best_copies $f_copies");
		push (@output, "$bench: $tune best        $f_best");
		push (@output, "$bench: $tune index       $f_index");
		for (@{$results->{$tune}{$bench}{'times'}}) {
		    $f_time = $_->[0][0] + $_->[0][1]/1000000;
		    push (@output, "$bench: $tune time      " . $_->[0][0] . " " .  $_->[0][1]);
		    for (@{$_->[1]}) {
			push (@output, "$bench: $tune messages $_");
		    }
		    push (@output, "$bench: $tune copies    " . $_->[2]);
		}
	    } else {
		if (!ref $results->{$tune}{$bench}) {
		    push (@output, "$bench: $tune " . $results->{$tune}{$bench});
		}
	    }
	}
    }
    for (@'extra_info, "metric", "input", "mode") {
	push (@output, "$_: " . $results->{$_});
    }

    if (@{$results->{'hardware'}}) {
	push (@output, "-HARDWARE-");
	for (@{$results->{'hardware'}}) {
	    if (!defined $_->[0]) {
		push(@output, sprintf (": %s", $_->[1]));
	    } else {
		$temp = $_->[0];
		$temp =~ s/ /_/g;
		push(@output, sprintf ("%s %s", $temp, $_->[1]));
	    }
	}
	push (@output, "");
	push (@output, "");
    }

    if (@{$results->{'software'}}) {
	push (@output, "-SOFTWARE-");
	for (@{$results->{'software'}}) {
	    if (!defined $_->[0]) {
		push(@output, sprintf (": %s", $_->[1]));
	    } else {
		$temp = $_->[0];
		$temp =~ s/ /_/g;
		push(@output, sprintf ("%s %s", $temp, $_->[1]));
	    }
	}
	push (@output, "");
	push (@output, "");
    }

    if (@{$results->{'notes'}}) {
	push (@output, "-NOTES-");
	$num = 0;
	for (@{$results->{'notes'}}) {
	    push (@output, sprintf ("note%03d: %s", $num++, $_));
	}
    }

    if (@{$results->{'errors'}}) {
	push (@output, "-ERRORS-");
	$num = 0;
	for (@{$results->{'errors'}}) {
	    push (@output, sprintf ("error%03d: %s", $num++, $_));
	}
    }

    @output;
}

1;
